from mot.model_building.model_functions import SimpleModelCLFunction, SimpleSampleModelCLPrototype
from mot.model_building.parameters import FreeParameter
from mot.library_functions import LogBesseli0
from mot.model_building.parameter_functions.transformations import ClampTransform


__author__ = 'Robbert Harms'
__date__ = "2014-08-05"
__license__ = "LGPL v3"
__maintainer__ = "Robbert Harms"
__email__ = "robbert.harms@maastrichtuniversity.nl"


class EvaluationModel(object):
    """The evaluation model is the model under which you evaluate your model estimates against observations.

    This class is a proxy for the optimization evaluation function and for the log likelihood objective function.

    This normally embed the noise model assumptions of your data.
    """

    @property
    def name(self):
        """Get the name of this model. Should be the same as the name in the CL header."""
        raise NotImplementedError()

    def get_noise_std_param_name(self):
        """Get the name of the parameter that is associated with the noise standard deviation in the problem data.

        Returns:
            str: the name of the parameter that is associated with the noise_std in the problem data.
        """
        raise NotImplementedError()

    def get_minimum_likelihood_function(self):
        """Get the function to evaluate the objective for a given observation and estimate under this noise model.

        This should return the observations such that when linearly summed we get the complete objective function value.

        Returns:
            mot.model_building.model_functions.SampleModelCLFunction: The objective function for the given observation
                index under this noise model.
        """
        raise NotImplementedError()

    def get_log_likelihood_function(self):
        """Get the function to evaluate the log likelihood for the given observations and model estimates.

        This should return the log likelihoods as such that when linearly summed they would yield the complete
        log likelihood for the model.

        Returns:
            mot.model_building.model_functions.SampleModelCLFunction: The log likelihood function for the given
                observation index under this noise model.
        """
        raise NotImplementedError()

    def get_prototype(self):
        """Get the cl function header for the evaluation model.

        It is assumed and required that both the minimum likelihood function as the log likelihood function
        have the same signature.

        Returns:
            mot.model_building.model_functions.SampleModelCLPrototype: the CL prototype for the CL functions in the
                the evaluation model.
        """
        raise NotImplementedError()


class SimpleAbstractEvaluationModel(EvaluationModel):

    def __init__(self, name, cl_function_name, parameter_list, noise_std_param_name=None,
                 dependency_list=()):
        """The evaluation model is the model under which you evaluate your model estimates against observations.

        This class is a proxy for the optimization evaluation function and for the log likelihood objective function.

        This normally embed the noise model assumptions of your data.

        Args:
            name (str): the name of this evaluation model
            cl_function_name (str): the name of the function, this is not used atm
            parameter_list (list or tuple): the list of parameters this model requires to function correctly
            noise_std_param_name (str): the name of the noise sigma parameter
            dependency_list (list or tuple): the list of function dependencies
        """
        self._name = name
        self._cl_function_name = cl_function_name
        self._parameter_list = parameter_list
        self._noise_std_param_name = noise_std_param_name
        self._dependency_list = dependency_list

    @property
    def name(self):
        return self._name

    def get_noise_std_param_name(self):
        return self._noise_std_param_name

    def get_minimum_likelihood_function(self):
        return SimpleModelCLFunction('double', self.name, self._cl_function_name, self._parameter_list,
                                     self._get_minimum_likelihood_code(),
                                     dependency_list=self._dependency_list)

    def get_log_likelihood_function(self):
        return SimpleModelCLFunction('double', self.name, self._cl_function_name, self._parameter_list,
                                     self._get_log_likelihood_code(),
                                     dependency_list=self._dependency_list)

    def get_prototype(self):
        return SimpleSampleModelCLPrototype('double', self.name, self._cl_function_name, self._parameter_list)

    def _get_minimum_likelihood_code(self):
        raise NotImplementedError()

    def _get_log_likelihood_code(self):
        raise NotImplementedError()


class SumOfSquaresEvaluationModel(SimpleAbstractEvaluationModel):

    def __init__(self):
        """Evaluates the distance between the estimated signal and the data using the sum of squared differences.

        This is implemented as:

        .. code-block:: c

            sum((observation - evaluation)^2)

        Since the optimization routines will sum the results, we only need to return:

        .. code-block:: c

            (observation - evaluation)^2

        And for sampling we return:

        .. code-block:: c

            -(observation - model_evaluation)^2
        """
        super(SumOfSquaresEvaluationModel, self).__init__(
            'SumOfSquares', 'sumOfSquares',
            [('double', 'observation'),
             ('double', 'model_evaluation')])

    def _get_minimum_likelihood_code(self):
        return '''
            return pown(observation - model_evaluation, 2);
        '''

    def _get_log_likelihood_code(self):
        return '''
            return -pown(observation - model_evaluation, 2);
        '''


class GaussianEvaluationModel(SimpleAbstractEvaluationModel):

    def __init__(self):
        """This uses the log of the Gaussian PDF for the maximum likelihood estimator and for the log likelihood.

        The PDF is defined as:

        .. code-block:: c

            PDF = 1/(sigma * sqrt(2*pi)) * exp(-(observation - evaluation)^2 / (2 * sigma^2))

        To have the joined probability over all instances one would normally have to take the product
        over all ``n`` instances:

        .. code-block:: c

            product(PDF)

        Instead of taking the product of this PDF we take the sum of the log of the PDF:

        .. code-block:: c

            sum(log(PDF))

        Where the log of the PDF is given by:

        .. code-block:: c

            log(PDF) = - ((observation - evaluation)^2 / (2 * sigma^2)) - log(sigma * sqrt(2*pi))

        We use this directly for sampling. For the optimization routines we need to adapt since the optimizers
        are minimization routines, we need use the negative, and since the optimizers already sum the results, we
        return here:

        .. code-block:: c

            = - log(PDF)
        """
        super(GaussianEvaluationModel, self).__init__(
            'GaussianNoiseModel', 'gaussianNoise',
            [('double', 'observation'),
             ('double', 'model_evaluation'),
             FreeParameter('mot_float_type', 'sigma', True, 1, 0, 'INFINITY', parameter_transform=ClampTransform())],
            noise_std_param_name='sigma')

    def _get_minimum_likelihood_code(self):
        """Get the Gaussian objective function.

        This omits the constant terms for speed reasons. Omitted terms are:

         .. code-block:: c

            + log(sigma * sqrt(2 * M_PI))
        """
        return '''
            return pown(observation - model_evaluation, 2) / (2 * sigma * sigma);
        '''

    def _get_log_likelihood_code(self):
        return '''
            return - pown(observation - model_evaluation, 2) / (2 * sigma * sigma)
                    - log(sigma * sqrt(2 * M_PI));
        '''


class OffsetGaussianEvaluationModel(SimpleAbstractEvaluationModel):

    def __init__(self):
        """This uses the log of the Gaussian PDF for the maximum likelihood estimator and for the log likelihood.

        The PDF is defined as:

        .. code-block:: c

            PDF = 1/(sigma * sqrt(2*pi)) * exp(-(observation - sqrt(evaluation^2 + sigma^2))^2 / (2 * sigma^2))

        To have the joined probability over all instances one would have to take the product over all n instances:

        .. code-block:: c

            product(PDF)

        Instead of taking the product of this PDF we take the sum of the log of the PDF:

        .. code-block:: c

            sum(log(PDF))

        Where the log of the PDF is given by:

        .. code-block:: c

            log(PDF) = - ((observation - sqrt(evaluation^2 + sigma^2))^2 / (2 * sigma^2)) - log(sigma * sqrt(2*pi))

        Which is used directly for sampling. For the maximum likelihood estimator we need to use the negative of
        this since the optimization routines are minimization routines, and since the optimization routines
        sum the results, we return here as MLE objective function:

        .. code-block:: c

            -log(PDF)
        """
        super(OffsetGaussianEvaluationModel, self).__init__(
            'OffsetGaussianNoise', 'offsetGaussian',
            [('double', 'observation'),
             ('double', 'model_evaluation'),
             FreeParameter('mot_float_type', 'sigma', True, 1, 0, 'INFINITY', parameter_transform=ClampTransform())],
            noise_std_param_name='sigma')

    def _get_minimum_likelihood_code(self):
        """Get the Offset Gaussian objective function.

        This omits the constant terms for speed reasons. Omitted terms are:

         .. code-block:: c

            (+ log(sigma * sqrt(2 * M_PI)))
        """
        return '''
            return pown(observation - hypot(model_evaluation, (double)sigma), 2) / (2 * sigma * sigma);
        '''

    def _get_log_likelihood_code(self):
        return '''
            double estimate = hypot(model_evaluation, (double)sigma);
            return - pown(observation - estimate, 2) / (2 * (sigma * sigma))
                    - log(sigma * sqrt(2 * M_PI));
        '''


class RicianEvaluationModel(SimpleAbstractEvaluationModel):

    def __init__(self):
        """This uses the log of the Rice PDF for the maximum likelihood estimator and for the log likelihood.

        The PDF is defined as:

        .. code-block:: c

            PDF = (observation/sigma^2)
                    * exp(-(observation^2 + evaluation^2) / (2 * sigma^2))
                    * bessel_i0((observation * evaluation) / sigma^2)

        Where where ``bessel_i0(z)`` is the modified Bessel function of the first kind with order zero. To have the
        joined probability over all instances one would have to take the product over all n instances:

        .. code-block:: c

            product(PDF)

        Instead of taking the product of this PDF over all instances we take the sum of the log of the PDF:

        .. code-block:: c

            sum(log(PDF))

        Where the log of the PDF is given by:

        .. code-block:: c

            log(PDF) = log(observation/sigma^2)
                        - (observation^2 + evaluation^2) / (2 * sigma^2)
                        + log(bessel_i0((observation * evaluation) / sigma^2))

        For sampling we use this directly. For the maximum likelihood estimator we need to use the negative of
        this since the optimization routines are minimization routines, and since the optimization routines
        sum the results, we return here as MLE objective function:

        .. code-block:: c

            -log(PDF).
        """
        super(RicianEvaluationModel, self).__init__(
            'RicianNoise', 'ricianNoise',
            [('double', 'observation'),
             ('double', 'model_evaluation'),
             FreeParameter('mot_float_type', 'sigma', True, 1, 0, 'INFINITY', parameter_transform=ClampTransform())],
            noise_std_param_name='sigma',
            dependency_list=(LogBesseli0(),))

    def _get_minimum_likelihood_code(self):
        """Get the Rician objective function.

        This omits the constant terms for speed reasons. Omitted terms are:

         .. code-block:: c

            + log(observation/sigma^2)
            - ((observation * observation) / (2 * (sigma * sigma)))

        """
        return '''
            return ((model_evaluation * model_evaluation) / (2 * sigma * sigma)) 
                    - log_bessel_i0((observation * model_evaluation) / (sigma * sigma));
        '''

    def _get_log_likelihood_code(self):
        return '''
            return log(observation / (sigma * sigma))
                    - ((observation * observation) / (2 * sigma * sigma))
                    - ((model_evaluation * model_evaluation) / (2 * sigma * sigma))
                    + log_bessel_i0((observation * model_evaluation) / (sigma * sigma));
        '''



import numpy as np

__author__ = 'Robbert Harms'
__date__ = "2014-10-23"
__license__ = "LGPL v3"
__maintainer__ = "Robbert Harms"
__email__ = "robbert.harms@maastrichtuniversity.nl"


class ParameterSampleStatistics(object):

    def get_statistics(self, samples):
        """Get the statistics for this parameter.

        Args:
            samples (ndarray): The 2d matrix (v, s) with for v voxels, s samples.

        Returns:
            SamplingStatistics: an object containing the sampling statistics
        """
        raise NotImplementedError()


class GaussianFit(ParameterSampleStatistics):
    """Calculates the mean and the standard deviation of the given samples.

    The standard deviation is calculated with a degree of freedom of one, meaning we are returning the unbiased
    estimator.
    """

    def get_statistics(self, samples):
        return SamplingStatisticsContainer(np.mean(samples, axis=1), {'std': np.std(samples, axis=1, ddof=1)})


class CircularGaussianFit(ParameterSampleStatistics):

    def __init__(self, max_angle=np.pi, min_angle=0):
        """Compute the circular mean for samples in a range

        The minimum angle is set to 0, the maximum angle can be given.

        Args:
            max_angle (float): The maximum angle used in the calculations
            min_angle (float): The minimum wrapped angle
        """
        super(CircularGaussianFit, self).__init__()
        self.max_angle = max_angle
        self.min_angle = min_angle

    def get_statistics(self, samples):
        from mot.cl_routines.mapping.circular_gaussian_fit import CircularGaussianFit
        mean, std = CircularGaussianFit().calculate(samples, high=self.max_angle, low=self.min_angle)
        return SamplingStatisticsContainer(mean, {'std': std})


class TruncatedGaussianFit(ParameterSampleStatistics):

    def __init__(self, low, high):
        """Fits a truncated gaussian distribution on the given samples.

        This may return mean values outside the bounds.

        Args:
            low (float): the lower bound of the truncated gaussian
            high (float): the upper bound of the truncated gaussian
        """
        self._low = low
        self._high = high

    def get_statistics(self, samples):
        from mot.cl_routines.mapping.truncated_gaussian_fit import TruncatedGaussianFit as fitter
        mean, std = fitter().calculate(samples, high=self._high, low=self._low)
        return SamplingStatisticsContainer(mean, {'std': std})


class TruncatedGaussianFitClipped(TruncatedGaussianFit):

    def __init__(self, low, high):
        """Fits a truncated gaussian distribution on the given samples and clips the mean to the given bounds.

        Args:
            low (float): the lower bound of the truncated gaussian
            high (float): the upper bound of the truncated gaussian
        """
        super(TruncatedGaussianFitClipped, self).__init__(low, high)

    def get_statistics(self, samples):
        statistics = super(TruncatedGaussianFitClipped, self).get_statistics(samples)
        clipped_expected_value = np.clip(statistics.get_expected_value(), self._low, self._high)
        return SamplingStatisticsContainer(clipped_expected_value, statistics.get_additional_statistics())


class TruncatedGaussianFitModulus(TruncatedGaussianFit):

    def __init__(self, low, high, modulus):
        """Fits a truncated gaussian distribution on the given samples and applies the modulus on the mean.

        Args:
            low (float): the lower bound of the truncated gaussian
            high (float): the upper bound of the truncated gaussian
            modulus (float): wrap the mean around this modulus
        """
        super(TruncatedGaussianFitModulus, self).__init__(low, high)
        self._modulus = modulus

    def get_statistics(self, samples):
        statistics = super(TruncatedGaussianFitModulus, self).get_statistics(samples)
        modulus_expected_value = np.mod(statistics.get_expected_value(), self._modulus)
        return SamplingStatisticsContainer(modulus_expected_value, statistics.get_additional_statistics())


class SamplingStatistics(object):

    def get_expected_value(self):
        """Get the expected value (typically the mean) of the given dataset.

        Returns:
            ndarray: The point estimate for every voxel.
        """
        raise NotImplementedError()

    def get_additional_statistics(self):
        """Get additional statistics about the parameter distribution.

        This normally returns only a dictionary with a standard deviation map, but it can return more statistics
        if desired.

        Returns:
            dict: dictionary with additional statistics. Example: ``{'std': ...}``
        """
        raise NotImplementedError()


class SamplingStatisticsContainer(SamplingStatistics):

    def __init__(self, expected_value, additional_maps):
        """Simple container for storing the point estimate and the other maps.

        Args:
            expected_value (ndarray): the array with the expected value (mean)
            additional_maps (dict): the additional maps
        """
        self._expected_value = expected_value
        self._additional_maps = additional_maps

    def get_expected_value(self):
        return self._expected_value

    def get_additional_statistics(self):
        return self._additional_maps



from scipy.stats import truncnorm

from mot.cl_routines.mapping.run_procedure import RunProcedure
from ...utils import SimpleNamedCLFunction, KernelInputBuffer, KernelInputScalar
from ...cl_routines.base import CLRoutine
import numpy as np

__author__ = 'Robbert Harms'
__date__ = '2017-09-11'
__maintainer__ = 'Robbert Harms'
__email__ = 'robbert.harms@maastrichtuniversity.nl'
__licence__ = 'LGPL v3'


class TruncatedGaussianFit(CLRoutine):

    def calculate(self, samples, high=np.pi, low=0):
        """Calculate the truncated normal mean and standard deviation.

        This uses the first two moments of the truncated normal distribution for approximating the fit.
        Please note that the sample means may lay outside of the high and low boundaries.

        This is basically a CL version of ``truncnorm(samples, low, high)`` in ``scipy.stats``.

        Args:
            samples (ndarray): (p, s) array with for p problems s samples.
            low (float): the low boundary for circular mean range
            high (float): the high boundary for circular mean range

        Returns:
            tuple: mean and std arrays
        """
        double_precision = samples.dtype == np.float64

        np_dtype = np.float32
        if double_precision:
            np_dtype = np.float64

        means = np.zeros(samples.shape[0], dtype=np_dtype, order='C')
        stds = np.zeros(samples.shape[0], dtype=np_dtype, order='C')
        samples = np.require(samples, np_dtype, requirements=['C', 'A', 'O'])

        mu, mu2 = truncnorm.stats(low, high, **{'moments': 'mv'})

        all_kernel_data = {'samples': KernelInputBuffer(samples),
                           'means': KernelInputBuffer(means, is_readable=False, is_writable=True),
                           'stds': KernelInputBuffer(stds, is_readable=False, is_writable=True),
                           'nmr_samples': KernelInputScalar(samples.shape[1]),
                           'mu': KernelInputScalar(mu),
                           'mu2': KernelInputScalar(mu2),
                           }

        runner = RunProcedure(**self.get_cl_routine_kwargs())
        runner.run_procedure(self._get_wrapped_function(), all_kernel_data, samples.shape[0],
                             double_precision=double_precision)

        return all_kernel_data['means'].get_data(), all_kernel_data['stds'].get_data()

    def _get_wrapped_function(self):
        func = '''
            void compute(mot_data_struct* data){
                double variance = 0;
                double mean = 0;
                double delta;
                double value;
                
                for(uint i = 0; i < data->nmr_samples; i++){
                    value = data->samples[i];
                    delta = value - mean;
                    mean += delta / (i + 1);
                    variance += delta * (value - mean);
                }
                variance /= (data->nmr_samples);
                
                double s_hat = sqrt(variance / data->mu2);
                double l_hat = mean - s_hat * data->mu;
                
                if(!isfinite(l_hat)){
                    l_hat = 0;
                }
                if(!(isfinite(s_hat) && (0 < s_hat))){
                    s_hat = 1;
                }
                
                *(data->means) = l_hat;
                *(data->stds) = s_hat;
            }            
        '''
        return SimpleNamedCLFunction(func, 'compute')



import numpy as np
from mot.cl_routines.mapping.run_procedure import RunProcedure
from ...utils import KernelInputBuffer, SimpleNamedCLFunction
from ...cl_routines.base import CLRoutine


__author__ = 'Robbert Harms'
__date__ = "2014-02-05"
__license__ = "LGPL v3"
__maintainer__ = "Robbert Harms"
__email__ = "robbert.harms@maastrichtuniversity.nl"


class ResidualCalculator(CLRoutine):

    def calculate(self, model, parameters):
        """Calculate and return the residuals, that is the errors, per problem instance per data point.

        Args:
            model (AbstractModel): The model to calculate the residuals of.
            parameters (ndarray): The parameters to use in the evaluation of the model

        Returns:
            Return per voxel the errors (eval - data) per protocol item
        """
        np_dtype = np.float32
        if model.double_precision:
            np_dtype = np.float64

        nmr_inst_per_problem = model.get_nmr_inst_per_problem()
        nmr_problems = model.get_nmr_problems()

        residuals = np.zeros((nmr_problems, nmr_inst_per_problem), dtype=np_dtype, order='C')

        all_kernel_data = dict(model.get_kernel_data())
        all_kernel_data.update({
            'parameters': KernelInputBuffer(parameters),
            'residuals': KernelInputBuffer(residuals, is_readable=False, is_writable=True)
        })

        runner = RunProcedure(**self.get_cl_routine_kwargs())
        runner.run_procedure(self._get_wrapped_function(model, parameters), all_kernel_data, parameters.shape[0],
                             double_precision=model.double_precision, use_local_reduction=False)

        return all_kernel_data['residuals'].get_data()

    def _get_wrapped_function(self, model, parameters):
        residual_function = model.get_residual_per_observation_function()
        param_modifier = model.get_pre_eval_parameter_modifier()
        nmr_params = parameters.shape[1]

        func = ''
        func += residual_function.get_cl_code()
        func += param_modifier.get_cl_code()
        func += '''
            void compute(mot_data_struct* data){
                mot_float_type x[''' + str(nmr_params) + '''];
                for(uint i = 0; i < ''' + str(nmr_params) + '''; i++){
                    x[i] = data->parameters[i];
                }
                
                ''' + param_modifier.get_cl_function_name() + '''(data, x);
                
                for(uint i = 0; i < ''' + str(model.get_nmr_inst_per_problem()) + '''; i++){
                    data->residuals[i] = ''' + residual_function.get_cl_function_name() + '''(data, x, i);
                }
            }
        '''
        return SimpleNamedCLFunction(func, 'compute')



import pyopencl as cl
import numpy as np
from ...utils import get_float_type_def, KernelInputDataManager
from ...cl_routines.base import CLRoutine
from ...load_balance_strategies import Worker


__author__ = 'Robbert Harms'
__date__ = "2014-02-05"
__license__ = "LGPL v3"
__maintainer__ = "Robbert Harms"
__email__ = "robbert.harms@maastrichtuniversity.nl"


class CalculateModelEstimates(CLRoutine):

    def calculate(self, model, parameters):
        """Evaluate the model for every problem and every observation and return the estimates.

        This only evaluates the model at the given data points. It does not use the problem data to calculate
        objective values.

        Args:
            model (AbstractModel): The model to evaluate.
            parameters (ndarray): The parameters to use in the evaluation of the model

        Returns:
            ndarray: Return per problem instance the evaluation per data point.
        """
        np_dtype = np.float32
        if model.double_precision:
            np_dtype = np.float64

        nmr_inst_per_problem = model.get_nmr_inst_per_problem()

        parameters = np.require(parameters, np_dtype, requirements=['C', 'A', 'O'])

        nmr_problems = parameters.shape[0]
        evaluations = np.zeros((nmr_problems, nmr_inst_per_problem), dtype=np_dtype, order='C')

        workers = self._create_workers(lambda cl_environment: _EvaluateModelWorker(
            cl_environment, self.get_compile_flags_list(model.double_precision), model, parameters, evaluations))
        self.load_balancer.process(workers, nmr_problems)

        return evaluations


class _EvaluateModelWorker(Worker):

    def __init__(self, cl_environment, compile_flags, model, parameters, evaluations):
        super(_EvaluateModelWorker, self).__init__(cl_environment)

        self._model = model
        self._data_info = self._model.get_kernel_data()
        self._data_struct_manager = KernelInputDataManager(self._data_info)
        self._double_precision = model.double_precision
        self._evaluations = evaluations
        self._parameters = parameters

        self._all_buffers, self._evaluations_buffer = self._create_buffers()
        self._kernel = self._build_kernel(self._get_kernel_source(), compile_flags)

    def calculate(self, range_start, range_end):
        nmr_problems = range_end - range_start
        self._kernel.get_estimates(self._cl_run_context.queue, (int(nmr_problems), ), None,
                                   *self._all_buffers, global_offset=(int(range_start),))
        self._enqueue_readout(self._evaluations_buffer, self._evaluations, range_start, range_end)

    def _create_buffers(self):
        evaluations_buffer = cl.Buffer(self._cl_run_context.context,
                                       cl.mem_flags.WRITE_ONLY | cl.mem_flags.USE_HOST_PTR,
                                       hostbuf=self._evaluations)

        all_buffers = [cl.Buffer(self._cl_run_context.context,
                                 cl.mem_flags.READ_ONLY | cl.mem_flags.USE_HOST_PTR,
                                 hostbuf=self._parameters),
                       evaluations_buffer]

        for data in [self._data_info[key] for key in sorted(self._data_info)]:
            all_buffers.append(cl.Buffer(self._cl_run_context.context,
                                         cl.mem_flags.READ_ONLY | cl.mem_flags.COPY_HOST_PTR, hostbuf=data.get_data()))

        return all_buffers, evaluations_buffer

    def _get_kernel_source(self):
        eval_function_info = self._model.get_model_eval_function()
        param_modifier = self._model.get_pre_eval_parameter_modifier()

        nmr_params = self._parameters.shape[1]

        kernel_param_names = ['global mot_float_type* params', 'global mot_float_type* estimates']
        kernel_param_names.extend(self._data_struct_manager.get_kernel_arguments())

        kernel_source = '''
            #define NMR_INST_PER_PROBLEM ''' + str(self._model.get_nmr_inst_per_problem()) + '''
        '''
        kernel_source += get_float_type_def(self._model.double_precision)
        kernel_source += self._data_struct_manager.get_struct_definition()
        kernel_source += eval_function_info.get_cl_code()
        kernel_source += param_modifier.get_cl_code()
        kernel_source += '''
            __kernel void get_estimates(
                ''' + ",\n".join(kernel_param_names) + '''
                ){
                    ulong gid = get_global_id(0);
                    mot_data_struct data = ''' + self._data_struct_manager.get_struct_init_string('gid') + ''';

                    mot_float_type x[''' + str(nmr_params) + '''];
                    for(uint i = 0; i < ''' + str(nmr_params) + '''; i++){
                        x[i] = params[gid * ''' + str(nmr_params) + ''' + i];
                    }

                    global mot_float_type* result = estimates + gid * NMR_INST_PER_PROBLEM;
                    
                    ''' + param_modifier.get_cl_function_name() + '''(&data, x);
                    
                    for(uint i = 0; i < NMR_INST_PER_PROBLEM; i++){
                        result[i] = ''' + eval_function_info.get_cl_function_name() + '''(&data, x, i);
                    }
            }
        '''
        return kernel_source
